@@ -22,14 +22,28 @@
 #include <gtk/gtk.h>
 #include <cspi/spi.h>
 #include <libgnomeui/gnome-ui-init.h>
+#include <popt.h>
 
 #include "graphics.h"
 #include "child-listener.h"
+#include "accessible-listener.h"
 
-/* Java requires disabling lots of sanity checks :-( */
-#define JAVA_ENABLED
+extern int max_children;
+extern gboolean dontdisplaychildren;
+extern gboolean use_table_if;
+
+static const struct poptOption options[] = {
+ 	{ "maxchildren", 'm', POPT_ARG_INT, &max_children, 1,
+	  "Specify the maximum number of children of an Accessible to be returned; or maximum ch ildren in a row or column if table is specified.\n", "INT"},
+ 	{ "dontdisplaychildren", 'd', POPT_ARG_NONE, &dontdisplaychildren, 1,
+	  "Do not display children for object which has state MANAGES_DESCENDANTS\n", NULL},
+ 	{ "table", 't', POPT_ARG_NONE, &use_table_if, 1,
+	  "Use AccessibleTable_getAccessibleAt to get children of Accessible which implements AccessibleTable\n", NULL},
+	{NULL, '\0', 0, NULL, 0, NULL, NULL}
+};
 
 extern void poke (Accessible *accessible);
+extern Accessible *get_accessible_at_index (GtkListStore *list_store, int i);
 
 #define APP_COL_NAME   0
 #define APP_COL_DESCR  1
@@ -108,22 +122,19 @@ validate_up_down_linkage (Accessible *accessible)
 	Accessible *parent;
 
 	aa = Accessible_queryInterface (accessible, "IDL:Accessibility/Accessible:1.0");
-#ifndef JAVA_ENABLED	
+
 	g_assert (aa == accessible);
-#else
-        if (aa != accessible) {
-		g_message ("aa == accessible failed");
-		fprintf (stderr,"cmp %p %p\n", aa, accessible);
-	}
-#endif
+
+	Accessible_unref (aa);
 	child = Accessible_getChildAtIndex (accessible, 0);
 	if (!child)
 		return;
 
 	parent = Accessible_getParent (child);
-#ifndef JAVA_ENABLED
+
 	g_assert (parent == accessible);
-#endif
+	Accessible_unref (child);
+	Accessible_unref (parent);
 }
 
 static void
@@ -140,8 +151,49 @@ static void
 application_clicked (GtkButton *button,
 		     AppWindow *app_window)
 {
-	validate_up_down_linkage (app_window->selected);
-	poke (app_window->selected);
+	if (!app_window->selected)
+		gdk_beep ();
+	else {
+		validate_up_down_linkage (app_window->selected);
+		poke (app_window->selected);
+	}
+}
+
+Accessible *
+get_accessible_at_index (GtkListStore *list_store, int i)
+{
+	GtkTreeIter iter;
+	GtkTreeModel *model = GTK_TREE_MODEL (list_store);
+	Accessible *accessible;
+	gboolean ret;
+
+	ret = gtk_tree_model_iter_nth_child (model, &iter, NULL, i);
+	g_return_val_if_fail (ret, NULL);
+
+	gtk_tree_model_get (model, &iter,
+			    APP_COL_HANDLE, &accessible,
+			    -1);
+	return accessible;
+}
+
+static void
+list_store_clear (GtkListStore *list_store)
+{
+	GtkTreeIter iter;
+	GtkTreeModel *model = GTK_TREE_MODEL (list_store);
+	Accessible *accessible;
+	gboolean ret;
+
+	ret = gtk_tree_model_get_iter_first (model, &iter);
+
+	while (ret) {
+		gtk_tree_model_get (model, &iter,
+				    APP_COL_HANDLE, &accessible,
+				    -1);
+		Accessible_unref (accessible);
+		ret = gtk_tree_model_iter_next (model, &iter);
+	}
+	gtk_list_store_clear (list_store);
 }
 
 static void
@@ -155,7 +207,7 @@ populate_app_list (AppWindow *app_window)
 
 	apps = Accessible_getChildCount (desktop);
 
-	gtk_list_store_clear (app_window->list_store);
+	list_store_clear (app_window->list_store);
 	for (i = 0; i < apps; i++) {
 		char *name, *descr;
 		Accessible *child;
@@ -181,6 +233,8 @@ populate_app_list (AppWindow *app_window)
 		GTK_TREE_MODEL (app_window->list_store), &iter))
 		gtk_tree_selection_select_iter (
 			gtk_tree_view_get_selection (app_window->list_view), &iter);
+	else
+		app_window->selected = NULL;
 
 	Accessible_unref (desktop);
 }
@@ -192,6 +246,12 @@ app_list_changed_cb (ChildListener *listener,
 	populate_app_list (window);
 }
 
+static void
+window_destroyed (GtkObject *obj)
+{
+	gtk_main_quit ();
+}
+
 static void
 create_app_list (AppWindow *app_window)
 {
@@ -246,6 +306,7 @@ application_window (void)
 	GClosure  *closure;
 	GladeXML  *app_xml;
 	AppWindow *app_window = g_new0 (AppWindow, 1);
+	ChildListener *cl;
 
 	app_xml = get_glade_xml ();
 	app_window->app_xml = app_xml;
@@ -292,17 +353,30 @@ application_window (void)
 	g_object_watch_closure (
 		G_OBJECT (app_window->window),
 		closure);
+	cl = child_listener_create (app_window->list_store);
 	g_signal_connect_closure (
-		child_listener_get (),
+		cl,
 		"app_list_changed",
 		closure, FALSE);
 
+	g_signal_connect (app_window->window,
+			  "destroy",
+			  G_CALLBACK (window_destroyed),
+			  NULL);
+			  
 	gtk_main ();
 
-	gtk_widget_destroy (app_window->window);
+	if (GTK_IS_WIDGET (app_window->window))
+		gtk_widget_destroy (app_window->window);
 
+	list_store_clear (app_window->list_store);
 	g_object_unref (app_xml);
 	g_free (app_window);
+	g_object_unref (cl);
+	/*
+	 * Deregister listeners which have been registered.
+	 */
+	accessible_listener_shutdown ();
 }
 
 static gboolean
@@ -331,7 +405,7 @@ main (int argc, char **argv)
 		if (!fork ()) {
 			putenv ("GTK_MODULES=gail:atk-bridge");
 			execlp ("gtk-demo", "gtk-demo", NULL);
-			g_warning ("Foo");
+			g_warning ("Unable to exec gtk-demo; check your PATH environment variable");
 			_exit (0);
 		}
 	}
@@ -342,7 +416,10 @@ main (int argc, char **argv)
 
 	gnome_program_init ("at-poke", VERSION,
 			    LIBGNOMEUI_MODULE,
-			    argc, argv, NULL);
+			    argc, argv, 
+			    GNOME_PARAM_POPT_TABLE, options, NULL);
+	if (max_children < 0)
+		max_children = 0;
 
 	application_window ();
 
