@@ -1,3 +1,4 @@
+/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
 /*
  * poke.c: the poker / accessible tree view
  *
@@ -10,12 +11,9 @@
 /*
  * TODO:
  *    + relation sets
- *    + whack a vertical divider into the app.
  *    + add 'getAccessibleAtPoint' on
  *      toplevel whatnot - if possible
- *    + Table view
  *    + Text selection view
- *    + AccessibleHypertext
  *    + AccessibleHyperlink
  *    + editable text
  *    + sync text caret spinbutton with text entry field, and changes to
@@ -79,12 +77,19 @@ typedef struct {
 	GtkWidget    *selection_clear;
 	gulong        selection_changed_id;
 
+	GtkListStore *table_columns_store;
+	GtkListStore *table_rows_store;
+
 	gulong        update_id;
 	gulong        text_update_id;
 
+	GtkListStore *hypertext_store;
+	GtkTreeView  *hypertext_view;
+
 	Accessible   *selected;
 
 	AccessibleCoordType ctype;
+	gboolean      poker_died;
 } Poker;
 
 enum {
@@ -100,6 +105,7 @@ enum {
 	RELATION_PATH
 };
 
+
 static void
 relation_row_activated_cb (GtkTreeView       *tree_view,
 			   GtkTreePath       *path,
@@ -233,6 +239,7 @@ update_if_accessible (Poker *poker, Accessible *accessible)
 		{ SPI_STATE_FOCUSED, "FOCUSED" },
 		{ SPI_STATE_HORIZONTAL, "HORIZONTAL" },
 		{ SPI_STATE_ICONIFIED, "ICONIFIED" },
+		{ SPI_STATE_MANAGES_DESCENDANTS, "MANAGES_DESCENDANTS" },
 		{ SPI_STATE_MODAL, "MODAL" },
 		{ SPI_STATE_MULTI_LINE, "MULTI_LINE" },
 		{ SPI_STATE_MULTISELECTABLE, "MULTISELECTABLE" },
@@ -464,7 +471,7 @@ update_if_text (Poker *poker, AccessibleText *text)
 						   &endOffset);
 
 	attribute_label_string = g_strconcat ("<i>",
-		(strlen (g_strstrip (attributes)) ?
+		(attributes && strlen (g_strstrip (attributes)) ?
 		 attributes : "(none)"), "</i>", NULL);
 
 	gtk_label_set_markup (GTK_LABEL (poker->attribute_label),
@@ -667,6 +674,57 @@ update_if_value (Poker *poker, AccessibleValue *value)
 				  poker->value_changed_id);
 }
 
+static void
+update_if_table (Poker *poker, AccessibleTable *table)
+{
+	long int rows, cols;
+	char *txt;
+	GtkWidget *widget;
+	int i;
+
+	rows = AccessibleTable_getNRows (table);
+	cols = AccessibleTable_getNColumns (table);
+
+	widget = glade_xml_get_widget (poker->xml, "table_if_rows");
+	txt = g_strdup_printf ("%ld", rows);
+	gtk_label_set_text (GTK_LABEL (widget), txt);
+	g_free (txt);
+
+	widget = glade_xml_get_widget (poker->xml, "table_if_columns");
+	txt = g_strdup_printf ("%ld", cols);
+	gtk_label_set_text (GTK_LABEL (widget), txt);
+	g_free (txt);
+
+
+	gtk_list_store_clear (poker->table_rows_store);
+	for (i = 0; i < rows; i++) {
+		GtkTreeIter iter;
+		char *description =
+			AccessibleTable_getRowDescription (table, i);
+		int extent =
+			AccessibleTable_getRowExtentAt (table, i, 0);
+		gtk_list_store_append (poker->table_rows_store, &iter);
+		gtk_list_store_set (poker->table_rows_store, &iter,
+				    0, description,
+				    1, extent,
+				    -1);
+	}
+
+	gtk_list_store_clear (poker->table_columns_store);
+	for (i = 0; i < cols; i++) {
+		GtkTreeIter iter;
+		char *description =
+			AccessibleTable_getColumnDescription (table, i);
+		int extent =
+			AccessibleTable_getColumnExtentAt (table, 0, i);
+		gtk_list_store_append (poker->table_columns_store, &iter);
+		gtk_list_store_set (poker->table_columns_store, &iter,
+				    0, description,
+				    1, extent,
+				    -1);
+	}
+}
+
 static void
 update_if_image (Poker *poker, AccessibleImage *image)
 {
@@ -674,6 +732,7 @@ update_if_image (Poker *poker, AccessibleImage *image)
 	long int w, h;
 
 	txt = AccessibleImage_getImageDescription (image);
+
 	gtk_label_set_text (GTK_LABEL (poker->image_descr),
 			    txt && txt [0] != '\0' ? txt : "<no description>");
 	SPI_freeString (txt);
@@ -814,6 +873,30 @@ selection_all_clicked_cb (GtkButton *button, Poker *poker)
 	AccessibleSelection_unref (selection);
 }
 
+static void
+update_if_hypertext (Poker *poker, AccessibleHypertext *hypertext)
+{
+	long i;
+	long num_links;
+	GtkTreeIter iter;
+
+	num_links = AccessibleHypertext_getNLinks (hypertext);
+	gtk_list_store_clear (poker->hypertext_store);
+
+	for (i = 0; i < num_links; i++) {
+		AccessibleHyperlink *link;
+		long n_anchors;
+
+		link = AccessibleHypertext_getLink (hypertext, i);
+		n_anchors = AccessibleHyperlink_getNAnchors (link);
+		Accessible_unref (link);
+		gtk_list_store_append (poker->hypertext_store, &iter);	
+		gtk_list_store_set (poker->hypertext_store, &iter,
+				    0, i,
+                                    1, n_anchors,
+                                    -1);
+	}
+}
 
 static const struct {
 	const char *widget_name;
@@ -828,9 +911,10 @@ static const struct {
 	{ "poker_text_frame",          "IDL:Accessibility/Text:1.0",         update_if_text },
 	{ "poker_editable_text_frame", "IDL:Accessibility/EditableText:1.0", update_if_editable },
 	{ "poker_value_frame",         "IDL:Accessibility/Value:1.0",        update_if_value },
-	{ "poker_table_frame",         "IDL:Accessibility/Table:1.0",        NULL }, /* FIXME */
+	{ "poker_table_frame",         "IDL:Accessibility/Table:1.0",        update_if_table },
 	{ "poker_image_frame",         "IDL:Accessibility/Image:1.0",        update_if_image },
 	{ "poker_selection_frame",     "IDL:Accessibility/Selection:1.0",    update_if_selection },
+	{ "poker_hypertext_frame",     "IDL:Accessibility/Hypertext:1.0",    update_if_hypertext },
 	{ NULL, NULL, NULL }
 };
 
@@ -889,6 +973,7 @@ poker_tree_selection_changed (GtkTreeSelection *selection,
 	GtkTreeModel *model;
 	
 	if (!gtk_tree_selection_get_selected (selection, &model, &iter)) {
+		poker->selected = NULL;
 		return;
 	}
 
@@ -902,6 +987,14 @@ poker_tree_selection_changed (GtkTreeSelection *selection,
 	update_frame_visibility (poker, accessible);
 }
 
+static void
+poker_tree_row_collapsed (GtkTreeView *view, GtkTreeIter *iter, GtkTreePath *path, Poker *poker)
+{
+	if (!poker->selected) {
+		gtk_tree_selection_select_iter (gtk_tree_view_get_selection (view), iter);
+	}
+}
+
 static void
 refresh_clicked_cb (GtkButton *button, Poker *poker)
 {
@@ -1027,6 +1120,12 @@ init_accessible_tree (Poker *poker, Accessible *accessible)
 		G_CALLBACK (poker_tree_selection_changed),
 		poker);
 
+	g_signal_connect (
+		poker->tree_view,
+		"row-collapsed",
+		G_CALLBACK (poker_tree_row_collapsed),
+		poker);
+
 	if (gtk_tree_model_get_iter_root (
 		GTK_TREE_MODEL (poker->tree_model), &iter)) {
 		GtkTreePath *path;
@@ -1043,15 +1142,26 @@ init_accessible_tree (Poker *poker, Accessible *accessible)
 static void
 root_died_cb (ChildListener *listener, Accessible *root, Poker *poker)
 {
+	poker->poker_died = TRUE;
 	if (root == poker->root_node)
 		gtk_main_quit ();
 }
 
+static void
+window_destroyed (GtkObject *obj, Poker *poker)
+{
+	if (!poker->poker_died) {
+		poker->poker_died = TRUE;
+ 		gtk_main_quit ();
+	}
+}
+
 void
 poke (Accessible *accessible)
 {
 	Poker *poker;
 	GtkWidget *widget;
+	gboolean poker_died;
 
 	poker = g_new0 (Poker, 1);
 	
@@ -1107,23 +1217,28 @@ poke (Accessible *accessible)
 	}
 
 	{ /* action interface */
+		GtkTreeView *treeview;
+
 		poker->action_store = gtk_list_store_new (
 			3, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);
 
 		poker->action_view = widget = glade_xml_get_widget (
 			poker->xml, "action_if_list");
 
-		gtk_tree_view_set_model (GTK_TREE_VIEW (widget),
+		treeview = GTK_TREE_VIEW (widget);
+		gtk_tree_view_set_model (treeview,
 					 GTK_TREE_MODEL (poker->action_store));
+
 		gtk_tree_view_insert_column_with_attributes (
-			GTK_TREE_VIEW (widget), 0, "Name",
+			treeview, 0, "Name",
 			gtk_cell_renderer_text_new (), "text", 0, NULL);
 		gtk_tree_view_insert_column_with_attributes (
-			GTK_TREE_VIEW (widget), 1, "Descr",
-			gtk_cell_renderer_text_new (), "text", 0, NULL);
+			treeview, 1, "Descr",
+			gtk_cell_renderer_text_new (), "text", 1, NULL);
 		gtk_tree_view_insert_column_with_attributes (
-			GTK_TREE_VIEW (widget), 2, "Keybinding",
-			gtk_cell_renderer_text_new (), "text", 0, NULL);
+			treeview, 2, "Keybinding",
+			gtk_cell_renderer_text_new (), "text", 2, NULL);
+		
 		g_signal_connect (widget, "row_activated",
 				  G_CALLBACK (action_row_activated_cb),
 				  poker);
@@ -1195,9 +1310,35 @@ poke (Accessible *accessible)
 					  poker);
 	}
 
+	{ /* table interface */
+		poker->table_rows_store = gtk_list_store_new (2, G_TYPE_STRING, G_TYPE_INT);
+		widget = glade_xml_get_widget (poker->xml, "table_if_row_data");
+		gtk_tree_view_set_model (GTK_TREE_VIEW (widget),
+					 GTK_TREE_MODEL (poker->table_rows_store));
+		gtk_tree_view_insert_column_with_attributes (
+			GTK_TREE_VIEW (widget), 0, "Row Description",
+			gtk_cell_renderer_text_new (), "text", 0, NULL);
+		gtk_tree_view_insert_column_with_attributes (
+			GTK_TREE_VIEW (widget), 1, "Height",
+			gtk_cell_renderer_text_new (), "text", 1, NULL);
+
+		poker->table_columns_store = gtk_list_store_new (2, G_TYPE_STRING, G_TYPE_INT);
+		widget = glade_xml_get_widget (poker->xml, "table_if_column_data");
+		gtk_tree_view_set_model (GTK_TREE_VIEW (widget),
+					 GTK_TREE_MODEL (poker->table_columns_store));
+		gtk_tree_view_insert_column_with_attributes (
+			GTK_TREE_VIEW (widget), 0, "Column Description",
+			gtk_cell_renderer_text_new (), "text", 0, NULL);
+		gtk_tree_view_insert_column_with_attributes (
+			GTK_TREE_VIEW (widget), 1, "Width",
+			gtk_cell_renderer_text_new (), "text", 1, NULL);
+	}
+
 	{ /* image interface */
 		poker->image_descr = glade_xml_get_widget (
 			poker->xml, "image_if_description");
+		/* Prevent width of image description expanding too much.*/
+		gtk_widget_set_size_request (poker->image_descr, 100, -1);
 		poker->image_size  = glade_xml_get_widget (
 			poker->xml, "image_if_size");
 	}
@@ -1237,6 +1378,28 @@ poke (Accessible *accessible)
 			"clicked", G_CALLBACK (selection_all_clicked_cb), poker);
 	}
 
+	{ /* hypertext interface */
+		poker->hypertext_store = gtk_list_store_new (
+			2, G_TYPE_INT, G_TYPE_INT);
+
+		widget = glade_xml_get_widget (
+			poker->xml, "hypertext_if_links");
+		poker->hypertext_view = GTK_TREE_VIEW (widget); 
+
+		gtk_tree_view_set_model (GTK_TREE_VIEW (widget),
+					 GTK_TREE_MODEL (poker->hypertext_store));
+		gtk_tree_view_insert_column_with_attributes (
+			GTK_TREE_VIEW (widget), 0, "Link",
+			gtk_cell_renderer_text_new (), "text",
+			0, NULL);
+
+		gtk_tree_view_insert_column_with_attributes (
+			GTK_TREE_VIEW (widget), 1, "Anchors",
+			gtk_cell_renderer_text_new (), "text",
+			1, NULL);
+
+	}
+
 	{ /* main frame buttons */
 		g_signal_connect (
 			glade_xml_get_widget (
@@ -1290,6 +1453,12 @@ poke (Accessible *accessible)
 			"focus_out_event",
 			G_CALLBACK (poker_window_activate_cb),
 			poker);
+
+		g_signal_connect (
+			poker->window,
+			"destroy",
+			G_CALLBACK (window_destroyed),
+			poker);
 	}
 
 	init_accessible_tree (poker, accessible);
@@ -1299,8 +1468,12 @@ poke (Accessible *accessible)
 
 	child_listener_remove_root (child_listener_get (), poker->root_node);
 
-	gtk_widget_destroy (poker->window);
-	
+	if (GTK_IS_WIDGET (poker->window))
+		gtk_widget_destroy (poker->window);
+	g_object_unref (poker->tree_model);	
 	g_object_unref (poker->xml);
+	poker_died = poker->poker_died;
 	g_free (poker);
+	if (!poker_died)
+		gtk_main_quit();
 }
